import { getActualPoint, getClickPoint, createSvg, createElement } from '../annotation/utils';
import { onClickOutside, setCss, isMobileDevice } from '../ui_utils';
import copy from 'copy-to-clipboard';

export class ImageEditor {
  constructor({
    eventBus,
    contentContainer,
    container,
    pagePtr,
    editPagePtr,
    editAreaPtr,
    editAreaIndex,
    viewport,
    scale,
    messageHandler,
    newAdd,
    pageViewer
  }) {
    this.eventBus = eventBus
    this.contentContainer = contentContainer
    this.container = container
    this.pagePtr = pagePtr
    this.editPagePtr = editPagePtr
    this.editAreaPtr = editAreaPtr
    this.editAreaIndex = editAreaIndex
    this.viewport = viewport
    this.scale = scale
    this.messageHandler = messageHandler
    this.newAdd = newAdd
    this.pageViewer = pageViewer

    this.frame = null
    this.borderWidth = 2
    this.pointWidth = 6
    this.pointHeight = 6
    this.activeCharPlace = null
    this.selectedRects = null
    this.selectedCharRange = null
    this.state = 0 // 0 未选中；1 已选中；2 裁剪状态
    this.mouseDown = false
    this.ratio = window.devicePixelRatio || 1
    this.opacity = 1

    this.start = null
    this.end = null
    this.newStart = null
    this.newEnd = null
    this.offsetX = 0
    this.offsetY = 0
    this.startPoint = null
    this.endPoint = null
    this.composing = false // 组合输入状态

    this.isFirefox = navigator.userAgent.indexOf('Firefox') > -1
    this.mousedown = isMobileDevice ? 'touchstart' : 'mousedown'
    this.mouseup = isMobileDevice ? 'touchend' : 'mouseup'
    this.mousemove = isMobileDevice ? 'touchmove' : 'mousemove'
    this.textInput = isMobileDevice || this.isFirefox ? 'input' : 'textInput'

    this.onClick = this.handleClick.bind(this)
    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onKeydown = this.handleKeyDown.bind(this)
    // this.onTextInput = this.handleTextInput.bind(this)
    this.onHandlePropertyPanelChanged = this.handlePropertyPanelChanged.bind(this)

    this.render()
  }

  async render () {
    this.eventBus._on('imagePropertyChanged', this.onHandlePropertyPanelChanged)

    await this.updateCanvas()
    this.opacity = await this.messageHandler.sendWithPromise('GetImageTransparency', this.editAreaPtr)

    let frameContainer = createElement(
      'div',
      {
        position: 'absolute',
        left: this.rect.left + 'px',
        top: this.rect.top + 'px',
        width: this.rect.width + 'px',
        height: this.rect.height + 'px'
      },
      {
        class: 'frame-container',
        index: this.editAreaIndex
      }
    )
    this.frameContainer = frameContainer
    this.container?.append(this.frameContainer)

    let imageContainer = createElement(
      'div',
      {
        position: 'absolute',
        left: 0,
        top: 0,
        width: '100%',
        height: '100%',
        borderWidth: this.borderWidth + 'px',
        borderColor: '#1460F3'
      },
      {
        class: 'image-container',
      }
    )
    this.imageContainer = imageContainer
    this.frameContainer.append(this.imageContainer)

    this.imageContainer.addEventListener('click', this.onClick)
    this.imageContainer.addEventListener(this.mousedown, this.onMousedown)
    this.imageContainer.addEventListener(this.mouseup, this.onMouseup)

    const outerLine = createSvg('svg', {
      class: 'outerline'
    }, {
      position: 'absolute',
      left: this.rect.left - this.borderWidth + 'px',
      top: this.rect.top - this.borderWidth + 'px',
      width: this.rect.width + this.borderWidth + this.pointWidth / 2 + 'px',
      height: this.rect.height + this.borderWidth + this.pointHeight / 2 + 'px',
      zIndex: 2
    })

    this.outerLine = outerLine

    this.moveRect = createSvg(
      "rect",
      {
        class: "move",
        'data-id': "move",
        stroke: "#1460F3",
        'stroke-width': this.borderWidth,
        'fill-opacity': 0,
        width: this.rect.width - this.borderWidth,
        height: this.rect.height - this.borderWidth,
        x: this.pointWidth / 2,
        y: this.pointHeight / 2
      }
    );

    this.topLeftRect = createSvg(
      "rect",
      {
        class: "nw-resize",
        'data-id': "nw-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: 0,
        y: 0
      }
    );

    this.LeftRect = createSvg(
      "rect",
      {
        class: "w-resize",
        'data-id': "w-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: 0,
        y: (this.rect.height - this.pointHeight + this.borderWidth) / 2
      }
    );

    this.bottomLeftRect = createSvg(
      "rect",
      {
        class: "sw-resize",
        'data-id': "sw-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: 0,
        y: this.rect.height - this.pointHeight / 2 + this.borderWidth / 2
      }
    );

    this.bottomRect = createSvg(
      "rect",
      {
        class: "s-resize",
        'data-id': "s-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: (this.rect.width - this.pointWidth + this.borderWidth) / 2,
        y: this.rect.height - this.pointHeight / 2 + this.borderWidth / 2
      }
    );
    
    this.bottomRightRect = createSvg(
      "rect",
      {
        class: "se-resize",
        'data-id': "se-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: this.rect.width - this.pointWidth / 2 + this.borderWidth / 2,
        y: this.rect.height - this.pointHeight / 2 + this.borderWidth / 2
      }
    );

    this.rightRect = createSvg(
      "rect",
      {
        class: "e-resize",
        'data-id': "e-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: this.rect.width - this.pointWidth / 2 + this.borderWidth / 2,
        y: (this.rect.height - this.pointHeight + this.borderWidth) / 2
      }
    );

    this.topRightRect = createSvg(
      "rect",
      {
        class: "ne-resize",
        'data-id': "ne-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: this.rect.width - this.pointWidth / 2 + this.borderWidth / 2,
        y: 0
      }
    );

    this.topRect = createSvg(
      "rect",
      {
        class: "n-resize",
        'data-id': "n-resize",
        fill: "#1460F3",
        stroke: "none",
        width: this.pointWidth,
        height: this.pointHeight,
        x: (this.rect.width - this.pointWidth + this.borderWidth) / 2,
        y: 0
      }
    );

    this.outerLine.append(this.moveRect)
    this.outerLine.append(this.topLeftRect)
    this.outerLine.append(this.LeftRect)
    this.outerLine.append(this.bottomLeftRect)
    this.outerLine.append(this.bottomRect)
    this.outerLine.append(this.bottomRightRect)
    this.outerLine.append(this.rightRect)
    this.outerLine.append(this.topRightRect)
    this.outerLine.append(this.topRect)

    this.outerLine.addEventListener(this.mousedown, this.onMousedown)
    this.outerLine.addEventListener(this.mouseup, this.onMouseup)

    if (this.newAdd) {
      this.goEditing()
    }
  }

  getActualRect (viewport, s, frame) {
    const { left: x1, top: y1, right: x2, bottom: y2 } = frame
    
    const start = getActualPoint(
      {
        x: x1,
        y: y1
      },
      viewport,
      s
    )
    const end = getActualPoint(
      {
        x: x2,
        y: y2
      },
      viewport,
      s
    )
    return {
      start,
      end
    }
  }

  rectCalc (start, end, padding = 5) {
    return {
      top: (start.y < end.y ? start.y : end.y) - padding,
      left: (start.x < end.x ? start.x : end.x) - padding,
      width: Math.abs(end.x - start.x) + padding * 2,
      height: Math.abs(end.y - start.y) + padding * 2,
    }
  }
  
  getInitialPoint (start, end) {
    const s = this.scale
    const startX = start.x
    const startY = start.y
    const endX = end.x
    const endY = end.y

    const { width, height, rotation } = this.viewport;
    let x1, y1, x2, y2;
    if (rotation === 0) {
      x1 = startX / s;
      y1 = startY / s; 
      x2 = endX / s;
      y2 = endY / s;
    } else if (rotation === 90) {
      x1 = startY / s;
      y1 = startX / s;
      x2 = endY / s;
      y2 = endX / s;
    } else if (rotation === 180) {
      x1 =( width - startX) / s;
      y1 = startY / s;
      x2 = (width - endX) / s;
      y2 = endY / s;
    } else {
      x1 = (height - startY) / s
      y1 = (width - startX) / s
      x2 = (height - endY) / s
      y2 = (width - endX)  / s
    }
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    };
  }

  handleMouseDown (e) {
    if (e.button === 2) return // 右键点击不执行

    if (this.state === 1) {
      const operatorId = e.target.getAttribute('data-id')
    
      const { pageX, pageY } = getClickPoint(e)
      this.startState = {
        operator: operatorId,
        clickX: pageX,
        clickY: pageY,
      }

      document.addEventListener(this.mousemove, this.onMousemove)
      
    } else if (this.state === 2) {
      if (isMobileDevice) {
        const offsetX = e.changedTouches[0].clientX - this.pageViewer.div.getBoundingClientRect().left
        const offsetY = e.changedTouches[0].clientY - this.pageViewer.div.getBoundingClientRect().top

        this.startPoint = {
          x: offsetX / this.scale,
          y: offsetY / this.scale
        }
      } else {
        const offsetX = e.offsetX
        const offsetY = e.offsetY

        this.startPoint = {
          x: (offsetX + this.rect.left) / this.scale,
          y: (offsetY + this.rect.top) / this.scale
        }
      }

      this.imageContainer.addEventListener(this.mousemove, this.onMousemove)
    }

    document.addEventListener(this.mouseup, this.onMouseup)

    this.selectedCharRange = null
    this.mouseMoved = false
    this.mouseDown = true
  }

  async handleMouseUp (e) {
    if (this.contentContainer.selectedFrameIndex !== -1 && this.contentContainer.selectedFrameIndex !== this.editAreaIndex && this.state === 0) return

    if (!this.mouseDown) return
    this.mouseDown = false

    if (e.type === 'touchend') {
      document.body.style.overscrollBehavior = 'auto';
    }

    let flag = true
    if (this.state === 0) {
      this.state = 1
      flag = false

      this.contentContainer.selectedFrameIndex = this.editAreaIndex
    }

    if (this.state === 1) {
      // this.frameContainer.classList.remove('editing')

      const oldPoint = {
        start: this.start,
        end: this.end
      }
      const newPoint = {
        start: this.newStart,
        end: this.newEnd
      }

      const { pageX, pageY } = getClickPoint(e)
      if (!(pageX === this.startState?.clickX && pageY === this.startState?.clickY) && this.newStart && this.newEnd) {
        this.start = this.newStart
        this.end = this.newEnd
      }
      
      this.container.append(this.outerLine)

      if (this.mouseMoved) {
        const { start, end } = this.getInitialPoint(this.start, this.end)
        const rect = {
          left: start.x,
          top: start.y,
          right: end.x,
          bottom: end.y
        }

        await this.messageHandler.sendWithPromise('SetFrame', {
          pagePtr: this.pagePtr,
          editAreaPtr: this.editAreaPtr,
          rect
        })
        this.updateCanvas({oldPoint, newPoint})

      }
      // else if (!this.mouseMoved && flag) {
      //   this.state = 2
      //   if (isMobileDevice) {
      //     this.eventBus.dispatch('changeIsEditorPanelDisabled')
      //   } else {
      //     this.eventBus.dispatch('contentPropertyChange', { type: 'image', isOpen: true })
      //   }
      // }
    }
    
    if (this.state === 2) {
      this.outerLine.remove()
      this.frameContainer.classList.add('editing')

      let endPoint
      if (isMobileDevice) {
        const offsetX = e.changedTouches[0].clientX - this.pageViewer.div.getBoundingClientRect().left
        const offsetY = e.changedTouches[0].clientY - this.pageViewer.div.getBoundingClientRect().top

        endPoint = {
          x: offsetX / this.scale,
          y: offsetY / this.scale
        }
      } else {
        const offsetX = e.offsetX
        const offsetY = e.offsetY

        endPoint = {
          x: (offsetX + this.rect.left) / this.scale,
          y: (offsetY + this.rect.top) / this.scale
        }
      }
      
      if (this.mouseMoved) {
        endPoint = this.endPoint
        this.cursor.style.display = 'block'
      }
      if (!this.startPoint) {
        this.startPoint = endPoint
      }

      // this.textarea.focus()
      // this.textarea.addEventListener('keydown', this.onKeydown)
      // this.textarea.addEventListener(this.textInput, this.onTextInput)
    }

    this.frameContainer.classList.add('selected')

    if (this.state === 1) {
      onClickOutside([this.imageContainer, this.outerLine, document.querySelector('.editor-panel'), document.getElementById('propertyPanelButton')], this.handleOutside.bind(this))
    }
    
    document.removeEventListener(this.mousemove, this.onMousemove)
    document.removeEventListener(this.mouseup, this.onMouseup)

    this.imageContainer.removeEventListener(this.mousemove, this.onMousemove)
  }

  async handleMouseMove (e) {
    if (this.contentContainer.selectedFrameIndex !== -1 && this.contentContainer.selectedFrameIndex !== this.editAreaIndex && this.state === 0) return
    
    if (e.type === 'touchmove') {
      document.body.style.overscrollBehavior = 'none';
    }

    if (this.state === 1) {
      const { pageX, pageY } = getClickPoint(e)
      const start = this.start
      const end = this.end
      const startState = this.startState
      if (!startState) return
      if (startState.clickX === pageX && startState.clickY === pageY) return

      const { width, height } = this.viewport
      if (startState.operator === 'nw-resize') {
        let x1 = pageX - (startState.clickX - this.start.x);
        let y1 = pageY - (startState.clickY - this.start.y);
        x1 = Math.min(end.x - 5, x1)
        x1 = Math.max(0, x1)
        y1 = Math.min(end.y - 5, y1)
        y1 = Math.max(0, y1)

        this.updateOutline({
          start: { x: x1, y: y1 },
          end,
        });
      } else if (startState.operator === 'w-resize') {
        let x1 = pageX - (startState.clickX - this.start.x);
        x1 = Math.min(end.x - 5, x1)
        x1 = Math.max(0, x1)

        this.updateOutline({
          start: { x: x1, y: start.y },
          end,
        });
      } else if (startState.operator === 'sw-resize') {
        let x1 = pageX - (startState.clickX - this.start.x);
        let y1 = pageY - (startState.clickY - this.end.y);
        x1 = Math.min(end.x - 5, x1)
        x1 = Math.max(0, x1)
        y1 = Math.min(height, y1)
        y1 = Math.max(start.y + 5, y1)

        this.updateOutline({
          start: { x: x1, y: start.y },
          end: { x: end.x, y: y1 },
        });
      } else if (startState.operator === 's-resize') {
        let y1 = pageY - (startState.clickY - this.end.y);
        y1 = Math.min(height, y1)
        y1 = Math.max(start.y + 5, y1)
        this.updateOutline({
          start,
          end: { x: end.x, y: y1 },
        });
      } else if (startState.operator === 'se-resize') {
        let x = pageX - (startState.clickX - this.end.x);
        let y = pageY - (startState.clickY - this.end.y);
        x = Math.min(width, x)
        x = Math.max(start.x + 5, x)
        y = Math.min(height, y)
        y = Math.max(start.y + 5, y)
        this.updateOutline({
          start,
          end: { x, y },
        });
      } else if (startState.operator === 'e-resize') {
        let x = pageX - (startState.clickX - this.end.x);
        x = Math.min(width, x)
        x = Math.max(start.x + 5, x)
        this.updateOutline({
          start,
          end: { x, y: end.y },
        });
      } else if (startState.operator === 'ne-resize') {
        let x = pageX - (startState.clickX - this.end.x);
        let y = pageY - (startState.clickY - this.start.y);
        x = Math.min(width, x)
        x = Math.max(start.x + 5, x)
        y = Math.min(end.y - 5, y)
        y = Math.max(0, y)
        this.updateOutline({
          start: { x: start.x, y },
          end: { x, y: end.y },
        });
      } else if (startState.operator === 'n-resize') {
        let y = pageY - (startState.clickY - this.start.y);
        y = Math.min(end.y - 5, y)
        y = Math.max(0, y)
        this.updateOutline({
          start: { x: start.x, y },
          end,
        });
      } else if (startState.operator === 'move') {
        let x1 = pageX - (startState.clickX - this.start.x);
        let y1 = pageY - (startState.clickY - this.start.y);
        let x2 = pageX - (startState.clickX - this.end.x);
        let y2 = pageY - (startState.clickY - this.end.y);

        const rect = {
          width: Math.abs(start.x -end.x),
          height: Math.abs(start.y -end.y)
        }
        if (x1 < x2) {
          x1 = Math.max(0, x1)
          x1 = Math.min(width - rect.width, x1)
          x2 = Math.max(rect.width, x2)
          x2 = Math.min(width, x2)
        } else {
          x2 = Math.max(0, x2)
          x2 = Math.min(width - rect.width, x2)
          x1 = Math.max(rect.width, x1)
          x1 = Math.min(width, x1)
        }

        if (y1 < y2) {
          y1 = Math.max(0, y1)
          y1 = Math.min(height - rect.height, y1)
          y2 = Math.max(rect.height, y2)
          y2 = Math.min(height, y2)
        } else {
          y2 = Math.max(0, y2)
          y2 = Math.min(height - rect.height, y2)
          y1 = Math.max(rect.height, y1)
          y1 = Math.min(height, y1)
        }

        this.updateOutline({
          start: { x: x1, y: y1 },
          end: { x: x2, y: y2 },
        });
      }
    }

    this.mouseMoved = true
  }

  async handleClick () {
    if (this.contentContainer.selectedFrameIndex !== -1 && this.contentContainer.selectedFrameIndex !== this.editAreaIndex && this.state === 0) return

    if (!isMobileDevice) {
      this.updateOverview()
      this.eventBus.dispatch('contentPropertyChange', { type: 'image', isOpen: true, opacity: this.opacity * 100 })
    }
  }

  async handleOutside () {
    this.state = this.state === 2 ? 1 : this.state === 1 ? 0 : this.state

    this.frameContainer.classList.remove('editing')

    this.eventBus.dispatch('contentPropertyChange', { type: 'image', isOpen: false })

    this.cursor?.remove()
    this.cursor = null

    if (this.state === 0) {
      this.frameContainer.classList.remove('selected')

      this.outerLine.remove()

      this.contentContainer.selectedFrameIndex = -1
    }

    if (this.state === 1) {
      document.getElementById('propertyPanelButton').click()

      const text = await this.getText()
      if (text.length === 0) {
        await this.outerLine.remove()
        this.contentContainer.selectedFrameIndex = -1
        this.remove()
        return
      }

      this.container.append(this.outerLine)

      this.imageContainer.removeEventListener(this.mousemove, this.onMousemove)

      onClickOutside([this.imageContainer, this.outerLine, document.querySelector('.editor-panel'), document.getElementById('propertyPanelButton')], this.handleOutside.bind(this))
    }

    // this.textarea.removeEventListener('keydown', this.onKeydown)
    // this.textarea.removeEventListener(this.textInput, this.onTextInput)
  }

  async handleKeyDown (e) {
    let keyCode = e.keyCode || e.which

    const toolKey = [9, 16, 17, 18, 20, 27, 37, 38, 39, 40, 91, 93]
    let isToolKey = e.ctrlKey || e.altKey || e.metaKey || toolKey.includes(keyCode)

    this.isToolKey = isToolKey

    if (keyCode === 8 || keyCode === 46) { // 8 delete键，46 backspace键
      if (this.activeCharPlace.SectionIndex === 0
        && this.activeCharPlace.LineIndex === 0
        && this.activeCharPlace.RunIndex === 0
        && this.activeCharPlace.CharIndex === -1
        && !this.selectedCharRange
      ) return

      let newChar = {}
      if (this.selectedCharRange) {
        newChar = await this.getCharPlace('DeleteChars')
        this.selectedCharRange = null
      } else if (keyCode === 8) {
        newChar = await this.getCharPlace('BackSpaceChar')
      } else {
        newChar = await this.getCharPlace('DeleteChar')
      }
      this.activeCharPlace = newChar
      const oldRect = this.rect
      await this.updateCanvas(null, oldRect)
      this.clearSelectText()
      this.updateCursorLine()
      return
    }

    if (this.selectedCharRange && !isToolKey) {
      const newChar = await this.getCharPlace('DeleteChars')
      this.activeCharPlace = newChar
      this.selectedCharRange = null
      const oldRect = this.rect
      await this.updateCanvas(null, oldRect)
      this.updateCursorLine()
    }
  }

  async handleTextInput (e) {
    if (this.isToolKey && !this.isPasteKey) return

    let data = e.data

    if (this.isFirefox) {
      if (e.inputType === 'insertLineBreak') {
        data = '\n'
      } else if (this.composing || data === null) {
        return
      }
    }

    const newChar = await this.messageHandler.sendWithPromise('InsertText', {
      editAreaPtr: this.editAreaPtr,
      char: this.activeCharPlace,
      text: data
    })
    this.activeCharPlace = newChar

    this.saveEdit()
    this.updateCanvas()
  }

  // 获取rect
  async getRect () {
    this.frame = await this.messageHandler.sendWithPromise('GetFrame', {
      pagePtr: this.pagePtr,
      editAreaPtr: this.editAreaPtr
    })

    const { start, end } = this.getActualRect(
      this.viewport,
      this.scale,
      this.frame
    )
    this.start = start
    this.end = end

    const rect = this.rectCalc(start, end)
    this.rect = rect
  }

  // 更新canvas图
  async updateCanvas(whole, oldRect) {
    await this.getRect()

    if (this.frameContainer && this.outerLine) {
      this.updateOutline({
        start: this.start,
        end: this.end
      })

      setCss(this.frameContainer, {
        left: this.rect.left + 'px',
        top: this.rect.top + 'px',
        width: this.rect.width + 'px',
        height: this.rect.height + 'px'
      })
    }

    if (whole) {
      const { oldPoint, newPoint } = whole
      const wholeArea = this.getEntireArea(oldPoint, newPoint)

      const { start, end } = this.getActualRect(
        this.viewport,
        this.scale,
        wholeArea
      )
      let wholeAreaRect = this.rectCalc(start, end)

      const oldRect = this.rectCalc(oldPoint.start, oldPoint.end)
      const newWholeArea = this.getEntireArea(oldRect, this.rect)
      const newWholeAreaRect = {
        left: newWholeArea.left,
        top: newWholeArea.top,
        width: newWholeArea.right - newWholeArea.left,
        height: newWholeArea.bottom - newWholeArea.top
      }
      if (
        parseInt(wholeAreaRect.left) !== parseInt(newWholeArea.left) ||
        parseInt(wholeAreaRect.top) !== parseInt(newWholeArea.top) ||
        parseInt(wholeAreaRect.width) !== parseInt(newWholeArea.width) ||
        parseInt(wholeAreaRect.height) !== parseInt(newWholeArea.height)
      ) {
        wholeAreaRect = newWholeAreaRect
      }

      const imgRect = {
        left: parseInt(wholeAreaRect.left * this.ratio),
        top: parseInt(wholeAreaRect.top * this.ratio),
        right: parseInt(wholeAreaRect.left * this.ratio) + parseInt(wholeAreaRect.width * this.ratio),
        bottom: parseInt(wholeAreaRect.top * this.ratio) + parseInt(wholeAreaRect.height * this.ratio),
        width: parseInt(wholeAreaRect.width * this.ratio),
        height: parseInt(wholeAreaRect.height * this.ratio),
      }
      this.imgRect = imgRect

    } else if (oldRect) {
      const wholeAreaRect = this.getEntireArea(oldRect, this.rect)

      const imgRect = {
        left: parseInt(wholeAreaRect.left * this.ratio),
        top: parseInt(wholeAreaRect.top * this.ratio),
        right: parseInt(wholeAreaRect.right * this.ratio),
        bottom: parseInt(wholeAreaRect.bottom * this.ratio),
        width: parseInt(wholeAreaRect.right * this.ratio) - parseInt(wholeAreaRect.left * this.ratio),
        height: parseInt(wholeAreaRect.bottom * this.ratio) - parseInt(wholeAreaRect.top * this.ratio),
      }
      this.imgRect = imgRect

    } else {
      const imgRect = {
        left: parseInt(this.rect.left * this.ratio),
        top: parseInt(this.rect.top * this.ratio),
        right: parseInt(this.rect.left * this.ratio) + parseInt(this.rect.width * this.ratio),
        bottom: parseInt(this.rect.top * this.ratio) + parseInt(this.rect.height * this.ratio),
        width: parseInt(this.rect.width * this.ratio),
        height: parseInt(this.rect.height * this.ratio),
      }
      this.imgRect = imgRect
    }

    let { imageArray } = await this.messageHandler.sendWithPromise('PushRenderTask', {
      pagePtr: this.pagePtr,
      scale: this.scale * this.ratio,
      left: this.imgRect.left,
      right: this.imgRect.right,
      bottom: this.imgRect.bottom,
      top: this.imgRect.top
    })
    this.imageArray = imageArray

    this.drawCanvas()
    this.saveEdit()
  }

  // 绘制canvas
  drawCanvas () {
    if (!this.imgRect) {
      this.updateCanvas()
      return
    }
    const pageCtx = this.pageViewer.canvas.getContext('2d')
    const imageData = pageCtx.createImageData(parseInt(this.imgRect.width), parseInt(this.imgRect.height))
    imageData.data.set(this.imageArray)
    pageCtx.putImageData(imageData, this.imgRect.left, this.imgRect.top)
  }

  // 更新光标位置
  async updateCursorLine () {
    let cursorPoints = await this.messageHandler.sendWithPromise('GetTextCursorPoints', {
      pagePtr: this.pagePtr,
      editAreaPtr: this.editAreaPtr,
      char: this.activeCharPlace
    })
    
    this.cursor.innerHTML = ''
    const cursorLine = createSvg(
      "line",
      {
        x1: cursorPoints.HighPointX * this.scale - this.rect.left,
        x2: cursorPoints.LowPointX * this.scale - this.rect.left,
        y1: cursorPoints.HighPointY * this.scale - this.rect.top,
        y2: cursorPoints.LowPointY * this.scale - this.rect.top,
        stroke: 'currentcolor'
      },
    )
    this.cursorLine = cursorLine
    this.cursor.append(cursorLine)
  }

  // 保存编辑
  saveEdit () {
    this.messageHandler.sendWithPromise('EndEdit', this.editPagePtr)
  }

  // 更新outerline框的位置大小
  updateOutline ({ start, end }) {
    const rect = this.rectCalc(start, end)

    this.newStart = start
    this.newEnd = end

    setCss(this.outerLine, {
      left: rect.left - this.borderWidth + 'px',
      top: rect.top - this.borderWidth + 'px',
      width: rect.width + this.borderWidth + this.pointWidth / 2 + 'px',
      height: rect.height + this.borderWidth + this.pointHeight / 2 + 'px',
    })

    this.moveRect.setAttribute('width', rect.width - this.borderWidth)
    this.moveRect.setAttribute('height', rect.height - this.borderWidth)
    
    this.LeftRect.setAttribute("y", (rect.height - this.pointHeight + this.borderWidth) / 2)

    this.bottomLeftRect.setAttribute("y", rect.height - this.pointHeight / 2 + this.borderWidth / 2)

    this.bottomRect.setAttribute("x", (rect.width - this.pointWidth + this.borderWidth) / 2)
    this.bottomRect.setAttribute("y", rect.height - this.pointHeight / 2 + this.borderWidth / 2)


    this.bottomRightRect.setAttribute("x", rect.width - this.pointWidth / 2 + this.borderWidth / 2)
    this.bottomRightRect.setAttribute("y", rect.height - this.pointHeight / 2 + this.borderWidth / 2)


    this.rightRect.setAttribute("x", rect.width - this.pointWidth / 2 + this.borderWidth / 2)
    this.rightRect.setAttribute("y", (rect.height - this.pointHeight + this.borderWidth) / 2)

    
    this.topRightRect.setAttribute("x", rect.width - this.pointWidth / 2 + this.borderWidth / 2)

    this.topRect.setAttribute("x", (rect.width - this.pointWidth + this.borderWidth) / 2)
  }

  async setProperty(props) {
    for (const item in props) {

      if (item === 'alignType') {
        if (!this.selectedCharRange) {
          await this.messageHandler.sendWithPromise('SetTextAligningSection', {
            editAreaPtr: this.editAreaPtr,
            alignType: props.alignType,
            char: activeCharPlace = {
              SectionIndex: 0,
              LineIndex: 0,
              RunIndex: 0,
              CharIndex: -1
            }
          })
        } else {
          await this.messageHandler.sendWithPromise('SetTextAligningRange', {
            editAreaPtr: this.editAreaPtr,
            alignType: props.alignType,
            start: this.selectedCharRange.start,
            end: this.selectedCharRange.end,
          })
        }
      }

      if (!this.selectedCharRange && item !== 'alignType') {
        await this.selectAllText()
      }

      if (item === 'opacity') {
        await this.messageHandler.sendWithPromise('SetCharsFontTransparency', {
          editAreaPtr: this.editAreaPtr,
          start: this.selectedCharRange.start,
          end: this.selectedCharRange.end,
          opacity: props.opacity / 100
        })
      }

      if (item === 'fontSize') {
        await this.messageHandler.sendWithPromise('SetCharsFontSize', {
          editAreaPtr: this.editAreaPtr,
          start: this.selectedCharRange.start,
          end: this.selectedCharRange.end,
          fontSize: props.fontSize
        })
      }

      if (item === 'fontFamily') {
        await this.messageHandler.sendWithPromise('SetFontFromNativeTrueTypeFont', {
          editAreaPtr: this.editAreaPtr,
          start: this.selectedCharRange.start,
          end: this.selectedCharRange.end,
          fontFamily: props.fontFamily
        })
      }
    }

    const oldRect = this.rect
    await this.updateCanvas(null, oldRect)
  }

  // 属性面板 修改属性
  async handlePropertyPanelChanged (props) {
    if (this.state === 0) return
    console.log(props)

    for (const item in props) {
      if (item === 'rotate') {
        await this.messageHandler.sendWithPromise('RotateImage', {
          editAreaPtr: this.editAreaPtr,
          angle: props[item]
        })
      }

      if (item === 'flip') {
        if (props[item] === 'x') {
          await this.messageHandler.sendWithPromise('HorizontalMirrorImage', this.editAreaPtr)
        } else if (props[item] === 'y') {
          await this.messageHandler.sendWithPromise('VerticalMirrorImage', this.editAreaPtr)
        }
      }

      if (item === 'opacity' && this.opacity !== props.opacity / 100) {
        await this.messageHandler.sendWithPromise('SetImageTransparency', {
          editAreaPtr: this.editAreaPtr,
          opacity: props.opacity / 100
        })
        this.opacity = props.opacity / 100
      }

      if (item === 'tool' && props.tool === 'replace') {
        await this.uploadFile()
        .then(async (data) => {
          const { imageBase64, width, height } = data
          const rect = {
            top: this.frame.top,
            left: this.frame.left,
            width: this.frame.right - this.frame.left,
            height: this.frame.bottom - this.frame.top
          }

          const imageRatio = width / height
          const rectRatio = rect.width / rect.height
          let scaledWidth, scaledHeight

          if (imageRatio > rectRatio) {
            scaledWidth = rect.width
            scaledHeight = rect.width / imageRatio
          } else {
            scaledWidth = rect.height * imageRatio
            scaledHeight = rect.height
          }
          
          let left, right, bottom, top
          const centerPoint = {
            x: rect.left + rect.width / 2,
            y: rect.top + rect.height / 2
          }

          left = centerPoint.x - scaledWidth / 2
          right = centerPoint.x + scaledWidth / 2
          bottom = centerPoint.y + scaledHeight / 2
          top = centerPoint.y - scaledHeight / 2

          this.editAreaPtr = await this.messageHandler.sendWithPromise('ReplaceImageAreaByStream', {
            pagePtr: this.pagePtr,
            editPagePtr: this.editPagePtr,
            editAreaPtr: this.editAreaPtr,
            rect: {
              left,
              right,
              bottom,
              top
            },
            imageBase64
          })

          this.saveEdit()
        })
        .catch((error) => {
          console.error(error)
        })
      }

      if (item === 'tool' && props.tool === 'export') {
        const { imageArray } = await this.messageHandler.sendWithPromise('ExtractOriginalImage', this.editAreaPtr)
        console.log(imageArray)
      }
    }

    const oldRect = this.rect
    await this.updateCanvas(null, oldRect)
  }

  // 计算编辑区域大小变化后，影响的整个区域的位置
  getEntireArea (oldPoint, newPoint) {
    let area = null
    if (oldPoint.start && newPoint.start) {
      let oldP = this.getInitialPoint(oldPoint.start, oldPoint.end)
      let newP = this.getInitialPoint(newPoint.start, newPoint.end)

      let left = Math.min(oldP.start.x, newP.start.x)
      let top = Math.min(oldP.start.y, newP.start.y)
      let right = Math.max(oldP.end.x, newP.end.x)
      let bottom = Math.max(oldP.end.y, newP.end.y)

      area = {
        left,
        top,
        right,
        bottom
      }

    } else if (oldPoint.left && newPoint.left) {
      let left = Math.min(oldPoint.left, newPoint.left)
      let top = Math.min(oldPoint.top, newPoint.top)
      let right = Math.max(oldPoint.left + oldPoint.width, newPoint.left + newPoint.width)
      let bottom = Math.max(oldPoint.top + oldPoint.height, newPoint.top + newPoint.height)

      area = {
        left,
        top,
        right,
        bottom
      }
    }
    
    return area
  }

  // 直接进入编辑模式
  goEditing () {
    this.state = 1
    this.contentContainer.selectedFrameIndex = this.editAreaIndex
    this.frameContainer.classList.add('selected')
    this.container.append(this.outerLine)

    if (!isMobileDevice) {
      this.updateOverview()
      this.eventBus.dispatch('contentPropertyChange', { type: 'image', isOpen: true })
    }

    // this.textarea.focus()
    // this.textarea.addEventListener('keydown', this.onKeydown)
    // this.textarea.addEventListener(this.textInput, this.onTextInput)

    onClickOutside([this.imageContainer, this.outerLine, document.querySelector('.editor-panel'), document.getElementById('propertyPanelButton')], this.handleOutside.bind(this))
  }

  // 删除区域
  async remove () {
    this.frameContainer.remove()

    await this.messageHandler.send('RemoveEditAreaByIndex', {
      editPagePtr: this.editPagePtr,
      editAreaIndex: this.editAreaIndex
    })
    // await this.messageHandler.send('RemoveEditArea', {
    //   editPagePtr: this.editPagePtr,
    //   editAreaPtr: this.editAreaPtr
    // })
    
    this.contentContainer.removeEditor(this.editAreaIndex)
  }

  // 上传图片并获取宽高
  uploadFile () {
    return new Promise((resolve, reject) => {
      const fileInput = document.createElement('input')
      fileInput.type = 'file'
      fileInput.accept = '.jpg, .jpeg, .png, .bmp'

      fileInput.onchange = () => {
        const file = fileInput.files[0]

        if (file.size > 2 * 1024 * 1024) {
          reject('文件大小超过2M限制')
          fileInput.remove()
          window.$message.error('The file size exceeds the 2M limit.', {
            duration: 3000,
            // icon: () => h('img', { src: MessageError })
          })
          return
        }

        if (file) {
          const reader = new FileReader()
          reader.onload = function (e) {
            const image = new Image()
            image.onload = function () {
              resolve({
                // file,
                imageBase64: e.target.result,
                width: image.width,
                height: image.height
              })
            }
            image.onerror = function (error) {
              reject(error)
            }
            image.src = e.target.result
          }
          reader.onerror = function (error) {
            reject(error)
          }
          reader.readAsDataURL(file)
          fileInput.remove()
        }
      }
      fileInput.click()
    })
  }

  // 更新属性面板预览图
  async updateOverview() {
    const canvasWidth = Math.ceil(this.rect.width)
    const canvasHeight = Math.ceil(this.rect.height)

    const { imageArray } = await this.messageHandler.sendWithPromise('ExtractThumbImage', {
      editAreaPtr: this.editAreaPtr,
      width: canvasWidth,
      height: canvasHeight
    })

    const canvas = document.createElement('canvas')
    const context = canvas.getContext('2d')

    canvas.width = canvasWidth
    canvas.height = canvasHeight
    
    const imageData = context.createImageData(canvasWidth, canvasHeight)
    const imageDataArray = new Uint8ClampedArray(imageArray.buffer)
    imageData.data.set(imageDataArray)
    context.putImageData(imageData, 0, 0)
    const overviewImageSrc = canvas.toDataURL('image/png', 1)

    this.eventBus.dispatch('contentPropertyChange', { type: 'image', overviewImageSrc })
  }
}
